from logging import getLogger
from typing import Any

from ..base.client import BaseClient
from ..constants import (
    BASE_URL,
)
from ..utils.helpers import prepare_data
from .models import Customer, CustomerMetadata

logger = getLogger(__name__)


class CustomerClient(BaseClient):
    def create(
        self,
        customer: CustomerMetadata | dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> Customer:
        """creates a new customer using an or named arguments

        Args:
            customer (Optional[CustomerMetadata  |  dict], optional): You
                customer data, it can be a dict, an instance of
                `abacatepay.customers.CustomerMetadata`.

        Returns:
            Customer: An instance of the new customer.
        """
        logger.debug(f'Creating customer with URL: {BASE_URL}/customer/create')

        json_data = prepare_data(customer or kwargs, CustomerMetadata)
        response = self._request(
            f'{BASE_URL}/customer/create',
            method='POST',
            json=json_data,
        )
        return Customer(**response.json()['data'])

    def list(self) -> list[Customer]:
        logger.debug(f'Listing customers with URL: {BASE_URL}/customer/list')
        response = self._request(f'{BASE_URL}/customer/list', method='GET')
        return [Customer(**bill) for bill in response.json()['data']]



from typing import Annotated, Optional

from pydantic import AliasChoices, BaseModel, Field, StringConstraints


class CustomerMetadata(BaseModel):
    """Customer model

    Args:
        tax_id (str):  the customer identifier such as CPF or CNPJ.
        name (str): the customer name
        email (str): the customer email address
        cellphone: (str): the customer phone number.
    """

    tax_id: str = Field(
        serialization_alias='taxId',
        validation_alias=AliasChoices('taxId', 'tax_id'),
    )
    name: str
    email: str
    cellphone: str


class CustomerInline(BaseModel):
    """The customer model attached to other models

    Args:
        metadata (Customer): the metadata of the customer.
    """

    metadata: Optional[CustomerMetadata] = None


CustomerID = Annotated[str, StringConstraints(pattern=r'^cust_[A-Za-z0-9]+$')]


class Customer(CustomerInline):
    """
    Customer returned by API.

    Args:
        id (CustomerID): the customer unique ID in abacatepay.
        tax_id (str) the customer identification (CPF or CNPJ).
        email (str): the customer's email
        name (str): the customer's name
        cellphone (str): the customer's phone number
    """

    id: CustomerID

    @property
    def tax_id(self) -> str:
        """the customer identification (CPF or CNPJ)."""
        return self.metadata.tax_id

    @property
    def name(self) -> str:
        """the customer's name"""
        return self.metadata.name

    @property
    def email(self) -> str:
        """the customer's email"""
        return self.metadata.email

    @property
    def cellphone(self) -> str:
        """the customer's phone number"""
        return self.metadata.cellphone



from logging import getLogger
from typing import Any

from ..base.client import BaseClient
from ..constants import BASE_URL
from ..utils.helpers import prepare_data
from .models import PixQrCode, PixQrCodeIn, PixStatus

logger = getLogger(__name__)


class PixQrCodeClient(BaseClient):
    def create(self, data: PixQrCodeIn | dict[str, Any], **kwargs: Any) -> PixQrCode:
        """
        Create a new Pix QR Code.

        Args:
            amount (int): The amount to be paid in cents.
            expires_in (int, optional): The expiration time in seconds.
                Defaults to None.
            description (str, optional): A description for the Pix QR Code.
                Defaults to None.
            customer (CustomerMetadata | dict, optional): Customer information.
                Defaults to None.
        Returns:
            PixQrCode: The created Pix QR Code object.
        """
        json_data = prepare_data(data or kwargs, PixQrCodeIn)
        logger.debug('Creating Pix QR Code: %s', json_data)

        response = self._request(
            f'{BASE_URL}/pixQrCode/create',
            method='POST',
            json=json_data,
        )
        logger.debug('Pix QR Code created successfully: %s', response.json())
        return PixQrCode.model_validate(response.json()['data'])

    def check(self, id: str) -> PixStatus:
        """
        Get the status of a Pix QR Code.

        Args:
            ID (str): The unique identifier of the Pix QR Code.

        Returns:
            PixStatus: The status of the Pix QR Code.
        """
        logger.debug(f'Getting status for Pix QR Code ID: {id}')
        response = self._request(
            f'{BASE_URL}/pixQrCode/check?id={id}',
            method='GET',
        )
        return PixStatus.model_validate(response.json()['data'])

    def simulate(self, id: str, metadata: dict[str, Any] | None = None) -> PixQrCode:
        """
        Simulate a Pix QR Code.

        Args:
            id (str): The unique identifier of the Pix QR Code.
            metadata (dict, optional): Additional metadata for the simulation.
                Defaults to {}.

        Returns:
            PixQrCode: The simulated Pix QR Code object.
        """
        logger.debug(f'Simulating Pix QR Code ID: {id}')
        response = self._request(
            f'{BASE_URL}/pixQrCode/simulate-payment?id={id}',
            method='POST',
            json=metadata or {},
        )
        return PixQrCode.model_validate(response.json()['data'])



from typing import Any

from pydantic import BaseModel


def prepare_data(
    data: BaseModel | dict[str, Any],
    model_cls: type[BaseModel],
    *,
    use_alias: bool = True,
) -> dict[str, Any]:
    """It formats the data to be in compliance with the given model class to
        be sent to the API

    Args:
        data (BaseModel | dict): the user input data.
        model_cls (type[BaseModel]): the model that will dumps.
        use_alias (bool, optional): if `True` dumps by alias. Defaults to True.
    """
    if isinstance(data, model_cls):
        validated_data = data.model_dump(by_alias=use_alias)
        return {k: v for k, v in validated_data.items() if v != {}}

    elif isinstance(data, dict):
        validated_data = model_cls.model_validate(data).model_dump(by_alias=use_alias)
        return {k: v for k, v in validated_data.items() if v != {}}

    raise TypeError('Invalid data type.')



from typing import Annotated, Optional

from annotated_types import Ge
from pydantic import AliasChoices, BaseModel, Field


class Product(BaseModel):
    """A product to be appended to a billing when it's being created

    Args:
        external_id (str): the product ID in your application.
        name (str): the product name.
        description (Optional[str]): a detailed description of the product.
        quantity (int): the number of units of the given product (min: 1).
        price (int): the price of the product in cents (min: 100).
    """

    external_id: str = Field(
        validation_alias=AliasChoices('externalId', 'external_id'),
        serialization_alias='externalId',
    )
    name: str
    description: Optional[str] = None
    quantity: Annotated[int, Ge(1)]
    price: Annotated[int, Ge(100)]


class ProductInline(BaseModel):
    """Represents a product inside a Billing.

    Args:
        id (str): the product ID.
        external_id (str): the product ID in your application.
        quantity (int): the number of units of the given product.
    """

    id: str
    external_id: str = Field(
        validation_alias=AliasChoices('external_id', 'externalId'),
        serialization_alias='externalId',
    )
    quantity: Annotated[int, Ge(1)]



from .client import BillingClient
from .models import (
    Billing,
    BillingIn,
    BillingList,
    BillingMetadata,
)

__all__ = [
    'BillingClient',
    'Billing',
    'BillingList',
    'BillingIn',
    'BillingMetadata',
]



from typing import Any, Literal

import requests

from ..constants import USER_AGENT
from ..utils.exceptions import (
    APIConnectionError,
    APITimeoutError,
    raise_for_status,
)


class BaseClient:
    def __init__(self, api_key: str):
        self.__api_key = api_key

    def _request(
        self,
        url: str,
        method: Literal['GET', 'POST', 'PUT', 'PATCH', 'DELETE'] = 'GET',
        **kwargs: Any,
    ) -> requests.Response:
        request = requests.Request(
            method,
            url,
            headers={
                'Authorization': f'Bearer {self.__api_key}',
                'User-Agent': USER_AGENT,
            },
            **kwargs,
        )
        try:
            prepared_request = request.prepare()
            with requests.Session() as s:
                response = s.send(prepared_request)

            raise_for_status(response)
            return response

        except requests.exceptions.Timeout:
            raise APITimeoutError(request=request)

        except requests.exceptions.ConnectionError:
            raise APIConnectionError(message='Connection error.', request=request)



from http import HTTPStatus
from typing import Literal

import requests


class APIError(Exception):
    """The exception was raised due to an API error."""

    message: str
    request: str

    def __init__(self, message: str, request: requests.Request) -> None:
        super().__init__(message)
        self.message = """The exception was raised due to an API error."""
        self.request = request

    def __str__(self) -> str:
        if self.message:
            return f'{self.message}'
        return ''


class APIStatusError(APIError):
    """Raised when an API response has a status code of 4xx or 5xx."""

    response: requests.Response
    status_code: int

    def __init__(self, message: str = '', *, response: requests.Response) -> None:
        super().__init__(message, response.request)
        self.response = response
        self.status_code = response.status_code


class ForbiddenRequest(APIStatusError):
    """
    Means that the request was unsuccessful due to a forbidden request.
    Maybe your API key is wrong?
    """

    status_code: Literal[HTTPStatus.FORBIDDEN]

    def __init__(self, response: requests.Response, message: str = ''):
        super().__init__(message, response=response)
        self.message = (
            'Means that the request was unsuccessful due to a '
            'forbidden request. Maybe your API key is wrong?'
        )
        self.response = response
        self.status_code = HTTPStatus.FORBIDDEN

    def __str__(self) -> str:
        if self.message:
            return (
                f'{self.message} \n Status Code: {self.response.status_code}'
                f' | Response: {self.response.text}'
            )
        return str(self.response.content, 'utf-8')


class UnauthorizedRequest(APIStatusError):
    """
    Means that the request was unsuccessful due to a forbidden request.
    Maybe your API key doesn't have enought permissions
    """

    status_code: Literal[HTTPStatus.UNAUTHORIZED]

    def __init__(self, response: requests.Response, message: str = ''):
        super().__init__(message, response=response)
        self.message = (
            'Means that the request was unsuccessful due to a forbidden '
            "request. Maybe your API key doesn't have enought permissions"
        )
        self.response = response
        self.status_code = HTTPStatus.UNAUTHORIZED

    def __str__(self) -> str:
        if self.message:
            return (
                f'{self.message} | Status Code: {self.response.status_code}'
                f' | Response: {self.response.text}'
            )
        return str(self.response.content, 'utf-8')


class APIConnectionError(APIError):
    """The request was unsuccessful due to a connection error.
    Check your internet connection"""

    def __init__(
        self,
        *,
        message: str = (
            'The request was unsuccessful due to a connection error. Check your internet connection'
        ),
        request: requests.Request,
    ) -> None:
        super().__init__(message, request)


class APITimeoutError(APIConnectionError):
    """The request got timed out. You might try checking
    your internet connection."""

    def __init__(self, request: requests.Request) -> None:
        super().__init__(
            message='Request timed out. Check your internet connection',
            request=request,
        )


class BadRequestError(APIStatusError):
    """The request was unsuccessful due to a bad request.
    Maybe the request syntax is wrong"""

    status_code: Literal[HTTPStatus.BAD_REQUEST]

    def __init__(self, response: requests.Response) -> None:
        self.response = response
        self.status_code = HTTPStatus.BAD_REQUEST

    def __str__(self) -> str:
        return (
            f'The request was unsuccessful due to a bad request. '
            'Maybe the request syntax is wrong. Message error: '
            f'{self.response.json()}'
        )


class NotFoundError(APIStatusError):
    status_code: Literal[HTTPStatus.NOT_FOUND]

    def __init__(self, message: str = '', *, response: requests.Response) -> None:
        super().__init__(message, response=response)
        self.status_code = HTTPStatus.NOT_FOUND

    def __str__(self) -> str:
        return (
            'The request was unsuccessful due to a not found error. '
            f'Error status 404 | Requested URL: {self.response.url}'
        )


class InternalServerError(APIStatusError):
    """The request was unsuccessful due to an internal server error."""

    status_code: Literal[500] = 500

    def __init__(self, response: requests.Response) -> None:
        super().__init__(
            message=(
                'The request was unsuccessful due to an internal server error.'
                " It's not your fault, just try again later."
            ),
            response=response,
        )


def raise_for_status(response: requests.Response) -> None:
    code_exc_dict = {
        HTTPStatus.BAD_REQUEST: BadRequestError(response=response),
        HTTPStatus.UNAUTHORIZED: UnauthorizedRequest(response=response),
        HTTPStatus.FORBIDDEN: ForbiddenRequest(response=response),
        HTTPStatus.NOT_FOUND: NotFoundError(response=response),
        HTTPStatus.INTERNAL_SERVER_ERROR: InternalServerError(response=response),
    }

    code = response.status_code
    if code == HTTPStatus.OK:
        return

    if code not in code_exc_dict and code >= HTTPStatus.BAD_REQUEST:
        raise APIStatusError(message=response.text, response=response)

    raise code_exc_dict.get(
        response.status_code,
        APIError(message=response.text, request=response.request),
    )



from logging import getLogger
from typing import Any

from ..base.client import BaseClient
from ..constants import BASE_URL
from ..utils.helpers import prepare_data
from .models import Coupon, CouponIn

logger = getLogger(__name__)


class CouponClient(BaseClient):
    def create(self, data: CouponIn | dict[str, Any] | None = None, **kwargs: Any) -> Coupon:
        """
        Create a new coupon.

        Args:
            data (Coupon): an instance of `abacatepay.coupons.models.Coupon` a dict
                or the named params following the model schema.

        Keyword args:
            code (str): The unique code for the coupon.
            discount_kind (DISCOUNT_KINDS): The type of discount (e.g., percentage
                or fixed amount).
            discount (int): The value of the discount.
            notes (Optional[str]): A description or note about the coupon. Defaults
                to None.
                - Serialization alias: 'notes'
                - Example: "Cupom de desconto pro meu público"
            max_redeems (Optional[int]): The maximum number of times the coupon can
                be redeemed.
                Defaults to -1 for unlimited redemptions.
            metadata (Optional[dict]): Additional metadata for the coupon. Defaults
                to None.

        Returns:
            Coupon: The response with the coupon data.
        """
        json_data = prepare_data(data or kwargs, CouponIn)
        logger.debug('creating coupon: %s', json_data)

        response = self._request(
            f'{BASE_URL}/coupon/create',
            method='POST',
            json=json_data,
        )
        logger.debug(f'Response from coupon creation: {response.json()}')
        return Coupon.model_validate(response.json()['data'])

    def list(self) -> list[Coupon]:
        """
        List all coupons.
        Returns:
            list[Coupon]: A list of coupon objects.
        """
        logger.debug(f'Listing coupons with URL: {BASE_URL}/coupon/list')
        response = self._request(f'{BASE_URL}/coupon/list', method='GET')
        data = response.json().get('data', [])

        if not data:
            logger.warning('No coupons found in the response.')
            return []

        return [Coupon.model_validate(item) for item in data]



"""
The Python SDK for the AbacatePay API

Basic usage:
```python
import abacatepay
from abacatepay.products import Product

token = "<your api token>"
client = abacatepay.AbacatePay(token)

billing = client.billing.create(
    products=[
        Product(
            external_id="123",
            name="Teste",
            quantity=1,
            price=101,
            description="Teste"
        )
    ],
    return_url="https://abacatepay.com",
    completion_url="https://abacatepay.com"
)
print(billing.data.url)
# > https://abacatepay.com/pay/aaaaaaa
```

More examples found on https://docs.abacatepay.com/
"""

from .billings import BillingClient
from .coupons import CouponClient
from .customers import CustomerClient
from .pixQrCode import PixQrCodeClient


class AbacatePay:
    def __init__(self, api_key: str):
        self.billing = BillingClient(api_key)
        self.customers = CustomerClient(api_key)
        self.coupons = CouponClient(api_key)
        self.pixQrCode = PixQrCodeClient(api_key)



from typing import Any, Optional

from pydantic import (
    BaseModel,
    Field,
)

from ..constants import DISCOUNT_KINDS


class CouponIn(BaseModel):
    """
    Represents a coupon model for creation.

    Attributes:
        code (str): The unique code for the coupon.
        discount_kind (DISCOUNT_KINDS): The type of discount (e.g., percentage
            or fixed amount).
        discount (int): The value of the discount.
        notes (Optional[str]): A description or note about the coupon. Defaults
            to None.
            - Serialization alias: 'notes'
            - Example: "Cupom de desconto pro meu público"
        max_redeems (Optional[int]): The maximum number of times the coupon can
            be redeemed.
            Defaults to -1 for unlimited redemptions.
        metadata (Optional[dict]): Additional metadata for the coupon. Defaults
            to None.
    """

    code: str
    discount_kind: DISCOUNT_KINDS = Field(
        serialization_alias='discountKind',
        description='Type of discount (e.g., percentage or fixed amount)',
        examples=['PERCENTAGE', 'FIXED_AMOUNT'],
    )
    discount: int
    notes: Optional[str] = Field(
        None,
        serialization_alias='notes',
        description="Coupon's description",
        examples=['Cupom de desconto pro meu público'],
    )
    max_redeems: Optional[int] = Field(
        -1,
        description=(
            'Maximum number of times the coupon can be redeemed. Defaults to -1 for unlimited.'
        ),
        serialization_alias='maxRedeems',
    )
    metadata: dict[str, Any] = Field({}, description='Additional metadata for the coupon.')


class Coupon(BaseModel):
    """
    Represents a coupon model.

    Attributes:
        id (str): The unique identifier for the coupon.
        discount_kind (str): The type of discount (e.g., percentage or
            fixed amount).
        discount (int): The value of the discount.
        status (str): The current status of the coupon (e.g., ACTIVE, INACTIVE)
        notes (Optional[str]): A description or note about the coupon.
        max_redeems (int): The maximum number of times the coupon can be
            redeemed.
        redeems_count (int): The number of times the coupon has been redeemed.
        created_at (str): The timestamp when the coupon was created.
        updated_at (str): The timestamp when the coupon was last updated.
    """

    id: str
    discount_kind: str = Field(
        ...,
        alias='discountKind',
        description='Type of discount (e.g., percentage or fixed amount)',
    )
    discount: int
    status: str
    notes: str | None = Field(None, validation_alias='notes', description="Coupon's description")
    max_redeems: int = Field(
        -1,
        validation_alias='maxRedeems',
        description='Maximum number of times the coupon can be redeemed.',
    )
    redeems_count: int = Field(
        0,
        validation_alias='redeemsCount',
        description='Number of times the coupon has been redeemed.',
    )
    created_at: str = Field(
        ...,
        validation_alias='createdAt',
        description='Timestamp when the coupon was created.',
    )
    updated_at: str = Field(
        ...,
        validation_alias='updatedAt',
        description='Timestamp when the coupon was last updated.',
    )
    dev_mode: bool = Field(
        ...,
        validation_alias='devMode',
        description='Indicates if the coupon is in development mode.',
    )



from typing import Any

from pydantic import BaseModel, Field

from ..constants import PIX_QR_CODE_STATUS
from ..customers.models import CustomerMetadata


class PixQrCode(BaseModel):
    """
    A representation of a Pix QRCode.

    Attributes:
        id (str): Unique identifier of the Pix QRCode.
        amount (int): Amount to be paid.
        status (str): Information about the status of the Pix QRCode.
        dev_mode (bool): Environment in which the Pix QRCode was created.
        brcode (str): Copy-and-paste code of the Pix QRCode.
        brcode_base64 (str): Base64 image of the Pix QRCode.
        platform_fee (int): Platform fees.
        created_at (str): Creation date of the Pix QRCode.
        updated_at (str): Update date of the Pix QRCode.
        expires_at (str): Expiration date of the Pix QRCode.
    """

    id: str = Field(
        description='Unique identifier of the Pix QRCode.',
        examples=['pix_char_123456'],
    )
    amount: int = Field(description='Amount to be paid.', examples=[100])
    status: PIX_QR_CODE_STATUS = Field(
        description=(
            'Information about the status of the Pix QRCode.'
            'Options: PENDING, EXPIRED, CANCELLED, PAID, REFUNDED'
        ),
        examples=['PENDING'],
    )
    dev_mode: bool = Field(
        description='Environment in which the Pix QRCode was created.',
        examples=[True],
        validation_alias='devMode',
    )
    brcode: str = Field(
        description='Copy-and-paste code of the Pix QRCode.',
        examples=['00020101021226950014br.gov.bcb.pix'],
        validation_alias='brCode',
    )
    brcode_base64: str = Field(
        description='Base64 image of the Pix QRCode.',
        examples=['data:image/png;base64,iVBORw0KGgoAAA'],
        validation_alias='brCodeBase64',
    )
    platform_fee: int = Field(
        description='Platform fees.',
        examples=[80],
        validation_alias='platformFee',
    )
    created_at: str = Field(
        description='Creation date of the Pix QRCode.',
        examples=['2025-03-24T21:50:20.772Z'],
        validation_alias='createdAt',
    )
    updated_at: str = Field(
        description='Update date of the Pix QRCode.',
        examples=['2025-03-24T21:50:20.772Z'],
        validation_alias='updatedAt',
    )
    expires_at: str = Field(
        description='Expiration date of the Pix QRCode.',
        examples=['2025-03-25T21:50:20.772Z'],
        validation_alias='expiresAt',
    )


class PixStatus(BaseModel):
    """Represents the status of a Pix QRCode.

    Attributes:
        status (str): Information about the status of the Pix QRCode.
        expires_at (str): Expiration date of the Pix QRCode.
    """

    status: PIX_QR_CODE_STATUS = Field(
        description=(
            'Information about the status of the Pix QRCode. '
            'Options: PENDING, EXPIRED, CANCELLED, PAID, REFUNDED'
        ),
        examples=['PENDING'],
    )
    expires_at: str | None = Field(
        None,
        description='Expiration date of the Pix QRCode.',
        examples=['2025-03-25T21:50:20.772Z'],
    )


class PixQrCodeIn(BaseModel):
    """Represents a Pix QRCode model for creation.

    Attributes:
        amount (int): Amount to be paid in cents.
        expires_in (int | None): Expiration time in seconds. Defaults to None.
        description (str | None): A description for the Pix QR Code. Defaults
            to None.
        customer (dict[str, Any] | CustomerMetadata): Customer information.
            Optional.
    """

    amount: int = Field(
        description='Amount to be paid in cents.',
        examples=[100],
    )
    expires_in: int | None = Field(
        None,
        description='Expiration time in seconds. Defaults to None.',
        examples=[3600],
    )
    description: str | None = Field(
        None,
        description='A description for the Pix QR Code. Defaults to None.',
        examples=['Payment for services'],
    )
    customer: dict[str, Any] | CustomerMetadata = Field(
        {}, description='Customer information. Optional.'
    )
